// import 'dart:math';

import 'package:kenken/kenken.dart';
import 'package:test/test.dart';
import 'dart:convert';

void main() {
  group('Functions:', () {
    test('isUniq works right', () {
      expect(isUniq([1, 2, 3]), equals(true));
      expect(isUniq([1, 2, 1]), equals(false));
      expect(isUniq([1, 2, 2]), equals(false));
      expect(
          isUniq([
            1,
            2,
          ]),
          equals(true));
      expect(
          isUniq([
            1,
          ]),
          equals(true));
    });

    test('isApplicable works right', () {
      expect((Sum().isApplicable([1, 2, 3])), equals(true));
      expect(Diff().isApplicable([1, 2, 3]), equals(false));
      expect(Diff().isApplicable([2, 3]), equals(true));
      expect(Div().isApplicable([1, 2, 3]), equals(false));
      expect(Div().isApplicable([2, 3]), equals(false));
      expect(Div().isApplicable([1, 3]), equals(true));
      expect(Mul().isApplicable([1, 2, 3]), equals(true));
    });
  });

  group('Rule object', () {
    test('return cell to display rule in it', () {
      // Let create rule "5+" with such cells:
      // |-----------------------|
      // |      |        |       | # r - row
      // |------=================| # c - column
      // |      || r2,c2 | r2,c3 |
      // |---------------========|
      // |      || r3,c2 ||      |
      // |------==========-------|
      // In which cell to display rule "5+" ?
      var rule = Rule([RC(2, 2), RC(2, 3), RC(3, 2)], Sum(), 5);
      // Right answer is RC(2,2), because it is topmost and leftmost:
      var cellToDisplay = rule.coordinatesToDisplayTheRule;
      expect(cellToDisplay.row, equals(2));
      expect(cellToDisplay.col, equals(2));
    });
  });

  group('Kenken (filled manually):', () {
    Kenken kenken;

    setUp(() {
      kenken = Kenken(3);

      kenken
        // RC 1,1 shoul be equal 1:
        ..addRuleFor([RC(1, 1)], SingleValue(), 1)
        // Sum of cells 1,2 and 1,3 should be equal 5:
        ..addRuleFor([RC(1, 2), RC(1, 3)], Sum(), 5)
        // Diff between 2,1 and 3,1 should be 1
        // Note: Using kenken.RCList() instead of RC()
        ..addRuleFor(
            kenken.RCList([
              [2, 1],
              [3, 1]
            ]),
            Diff(),
            1)
        // sum of 2,2 2,3 should be 4
        ..addRuleFor([RC(2, 2), RC(2, 3)], Sum(), 4);

      // Do not add rule below to kenken now.
      // Let kenken will be not fully covered by custom reles.
      // This rule we will add later in test.
      // sum of 3,2 and 3,3 should be 3:
      // ..addRuleFor([RC(3, 2), RC(3, 3)], Sum(, label: '+'), 3);
    });

    test('`coordinatesOfRow` method returns right coordinates', () {
      var row2Coordinates = kenken.coordinatesOfRow(2);
      // first coordinate of row2Coordinates:
      expect(row2Coordinates[0].row, equals(2));
      expect(row2Coordinates[0].col, equals(1));
      // third coordinate of row2Coordinates:
      expect(row2Coordinates[2].row, equals(2));
      expect(row2Coordinates[2].col, equals(3));
    });

    test('rules not cover full kenken', () {
      expect(kenken.notCoveredByCustomRulesCells.length, equals(2));
    });

    test('rules cover full kenken', () {
      // Let add final rule.
      // sum of 3,2 and 3,3 should be 3:
      // Now the kenken is fully covered by custom rules.
      kenken.addRuleFor([RC(3, 2), RC(3, 3)], Sum(), 3);
      expect(kenken.notCoveredByCustomRulesCells.length, equals(0));
    });

    test('Right solved kenken has not invalid rules', () {
      var values = [
        [1, 2, 3], // row = 1,
        [2, 3, 1], // row = 2,
        [3, 1, 2], // row = 3,
      ];

      // Fill kenken cells with values as it does user:
      for (var r in kenken.rowOrColNumbers) {
        for (var c in kenken.rowOrColNumbers) {
          kenken.setValue(r, c, values[r - 1][c - 1]);
        }
      }

      // expect(awesome.isAwesome, isTrue);
      expect(kenken.brokenRules, equals([]));

      print('$kenken');
    });
  });

  group('Kenken (filled automatically):', () {
    Kenken kenken;

    setUp(() {
      kenken = Kenken(3);
    });

    test('Kenken auto fill values', () {
      var autoFilled = kenken.autoGeneratedValues;
      // print('auto filled values: $autoFilled');

      for (var r in kenken.rowOrColNumbers) {
        for (var c in kenken.rowOrColNumbers) {
          kenken.setValue(r, c, autoFilled[r - 1][c - 1]);
        }
      }

      expect(kenken.brokenRules, equals([]));
    });


    test('json', () {
      kenken.autoFillCustomRules();

      kenken.recording = true;

      kenken.setValue(1, 2, 5);
      expect(kenken.history.length, equals(1));
      kenken.setValue(1, 3, 4);
      expect(kenken.history.length, equals(2));

      var originKenken = kenken; // alias for short code
      var kenkenJsonString = jsonEncode(originKenken);
      print(kenkenJsonString);
      var kenkenFromJson = Kenken.fromJson(jsonDecode(kenkenJsonString));

      print(originKenken);
      print(kenkenFromJson);


      for (var r in originKenken.rowOrColNumbers) {
        for (var c in originKenken.rowOrColNumbers) {
          var cellKK1 = originKenken.cellOf(RC(r, c));
          var cellKK2 = kenkenFromJson.cellOf(RC(r, c));

          expect(cellKK1.value, equals(cellKK2.value));
          expect(cellKK1.ruleText, equals(cellKK2.ruleText));

          expect(cellKK1.itRule.cellsCoordinates[0].row, equals(cellKK2.itRule.cellsCoordinates[0].row));
          expect(cellKK1.itRule.cellsCoordinates[0].col, equals(cellKK2.itRule.cellsCoordinates[0].col));

          expect(cellKK1.itRule.operation.label, equals(cellKK2.itRule.operation.label));
          expect(cellKK1.itRule.operation.predicate, equals(cellKK2.itRule.operation.predicate));
          expect(cellKK1.itRule.validOperationResult, equals(cellKK2.itRule.validOperationResult));
        }
      }

      expect(originKenken.history.length, equals(kenkenFromJson.history.length));
      expect(originKenken.initState.length, equals(kenkenFromJson.initState.length));

    });

    test('history', (){
      kenken.autoFillCustomRules();
      kenken.recording = true;
      kenken.setValue(1, 1, 1);
      kenken.setValue(1, 2, 2);
      expect(kenken.history.length, equals(2));

      kenken.recording = false;
      kenken.setValue(1, 3, 3, recordIt: true);
      expect(kenken.history.length, equals(3));

      kenken.resetToInitState();
      for (var step in kenken.history){
        kenken.applyStep(step);
      }
      expect(kenken.history.length, equals(3));
    });
  });
}
